# flake8: noqa

from skale.contracts.base_contract import BaseContract, transaction_method

from skale.contracts.manager import Manager
from skale.contracts.contract_manager import ContractManager
from skale.contracts.token import Token
from skale.contracts.groups import Groups
from skale.contracts.constants_holder import ConstantsHolder

from skale.contracts.data.schains_data import SChainsData
from skale.contracts.data.nodes_data import NodesData
from skale.contracts.data.monitors_data import MonitorsData

from skale.contracts.functionality.schains import SChains
from skale.contracts.functionality.nodes import Nodes
from skale.contracts.functionality.monitors import MonitorsFunctionality

from skale.contracts.delegation.delegation_controller import DelegationController
from skale.contracts.delegation.validator_service import ValidatorService
from skale.contracts.delegation.token_state import TokenState
from skale.contracts.delegation.distributor import Distributor

from skale.contracts.dkg import DKG



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE manager operations """


import logging
import socket


from skale.contracts import BaseContract, transaction_method
from skale.utils import helper
from skale.utils.constants import GAS, OP_TYPES

logger = logging.getLogger(__name__)


class Manager(BaseContract):
    @transaction_method(GAS['create_node'])
    def create_node(self, ip, port, name, public_ip=None):
        logger.info(
            f'create_node: {ip}:{port}, public ip: {public_ip} name: {name}')
        skale_nonce = helper.generate_nonce()
        pk = self.skale.wallet.public_key

        if not public_ip:
            public_ip = ip

        transaction_data = self.create_node_data_to_bytes(
            ip, public_ip, port, name, pk, skale_nonce)
        return self.contract.functions.createNode(transaction_data)

    def create_node_data_to_bytes(self, ip, public_ip, port, name, pk, nonce):
        pk_fix = str(pk)[2:]

        type_bytes = OP_TYPES['create_node'].to_bytes(1, byteorder='big')
        port_bytes = port.to_bytes(2, byteorder='big')
        nonce_bytes = nonce.to_bytes(2, byteorder='big')  # todo
        ip_bytes = socket.inet_aton(ip)
        public_ip_bytes = socket.inet_aton(public_ip)
        pk_bytes = bytes.fromhex(pk_fix)
        name_bytes = name.encode()

        data_bytes = (
            type_bytes
            + port_bytes  # noqa: W503
            + nonce_bytes  # noqa: W503
            + ip_bytes  # noqa: W503
            + public_ip_bytes  # noqa: W503
            + pk_bytes  # noqa: W503
            + name_bytes  # noqa: W503
        )
        logger.info(
            f'create_node_data_to_bytes bytes: {self.skale.web3.toHex(data_bytes)}'
        )

        return data_bytes

    def create_default_schain(self, name):
        lifetime = 3600
        nodes_type = 4
        price_in_wei = self.skale.schains.get_schain_price(nodes_type, lifetime)
        return self.create_schain(lifetime, nodes_type, price_in_wei, name,
                                  wait_for=True)

    @transaction_method(GAS['create_schain'])
    def create_schain(self, lifetime, type_of_nodes, deposit, name):
        logger.info(
            f'create_schain: type_of_nodes: {type_of_nodes}, name: {name}')

        token = self.skale.get_contract_by_name('token')
        skale_nonce = helper.generate_nonce()
        transaction_data = self.create_schain_data_to_bytes(
            lifetime, type_of_nodes, name, skale_nonce)

        return token.contract.functions.send(self.address, deposit,
                                             transaction_data)

    def create_schain_data_to_bytes(self, lifetime, type_of_nodes, name,
                                    nonce):
        type_bytes = OP_TYPES['create_schain'].to_bytes(1, byteorder='big')

        lifetime_hex = lifetime.to_bytes(32, byteorder='big')
        type_of_nodes_hex = type_of_nodes.to_bytes(1, byteorder='big')
        nonce_hex = nonce.to_bytes(2, byteorder='big')
        name_hex = name.encode()

        data_bytes = (type_bytes + lifetime_hex +
                      type_of_nodes_hex + nonce_hex + name_hex)
        logger.info(
            f'create_schain_data_to_bytes bytes:'
            f'{self.skale.web3.toHex(data_bytes)}'
        )
        return data_bytes

    @transaction_method(GAS['get_bounty'])
    def get_bounty(self, node_id):
        return self.contract.functions.getBounty(node_id)

    @transaction_method(GAS['send_verdict'])
    def send_verdict(self, validator, node_id, downtime, latency):
        return self.contract.functions.sendVerdict(validator, node_id, downtime,
                                                   latency)

    @transaction_method(GAS['send_verdicts'])
    def send_verdicts(self, validator, nodes_ids, downtimes, latencies):
        return self.contract.functions.sendVerdicts(validator, nodes_ids,
                                                    downtimes, latencies)

    @transaction_method(GAS['delete_node'])
    def deregister(self, node_id):
        return self.contract.functions.deleteNode(node_id)

    @transaction_method(GAS['delete_schain'])
    def delete_schain(self, schain_name):
        return self.contract.functions.deleteSchain(schain_name)

    @transaction_method(GAS['delete_node_by_root'])
    def delete_node_by_root(self, node_id):
        return self.contract.functions.deleteNodeByRoot(node_id)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from web3 import Web3
from Crypto.Hash import keccak

from skale.schain_config import NODE_OWNER_ALLOC, SCHAIN_OWNER_ALLOC


def add_to_accounts(allocation, account, amount, code=None, storage={}, nonce=0):
    assert isinstance(code, str) or code is None
    assert isinstance(storage, dict) or storage is None
    acc_fx = Web3.toChecksumAddress(account)
    if str(acc_fx) not in allocation:
        allocation[acc_fx] = {"balance": str(amount)}
        if code:
            allocation[acc_fx]['code'] = code
            allocation[acc_fx]['storage'] = storage
            allocation[acc_fx]['nonce'] = str(nonce)


def update_accounts(schain, schain_nodes):
    allocation = {}
    add_to_accounts(allocation, schain['owner'], SCHAIN_OWNER_ALLOC)
    for node in schain_nodes:
        if str(node['owner']) not in allocation:
            add_to_accounts(allocation, node['owner'], NODE_OWNER_ALLOC)
    add_to_accounts(
        allocation,
        '0xD2001000000000000000000000000000000000D2',
        0,
        *_generate_context_predeployed_smart_contract(
            schain['owner'],
            schain['name']
        )
    )
    return allocation


def update_base_config(base_config, schain, schain_nodes):
    new_accounts = update_accounts(schain, schain_nodes)
    base_config['accounts'] = {**base_config['accounts'], **new_accounts}
    add_chain_id(base_config, schain['name'])


def _generate_context_predeployed_smart_contract(owner_address: str,
                                                 schain_name: str) -> (str, dict):
    code = ('0x608060405234801561001057600080fd5b506004361061005e576000357c0100' +
            '0000000000000000000000000000000000000000000000000000009004806315' +
            'cd7a2d1461006357806367091a01146100e657806383e781fe1461012a575b60' +
            '0080fd5b61006b610174565b6040518080602001828103825283818151815260' +
            '200191508051906020019080838360005b838110156100ab5780820151818401' +
            '52602081019050610090565b50505050905090810190601f1680156100d85780' +
            '820380516001836020036101000a031916815260200191505b50925050506040' +
            '5180910390f35b610128600480360360208110156100fc57600080fd5b810190' +
            '80803573ffffffffffffffffffffffffffffffffffffffff1690602001909291' +
            '90505050610216565b005b6101326102b2565b604051808273ffffffffffffff' +
            'ffffffffffffffffffffffffff1673ffffffffffffffffffffffffffffffffff' +
            'ffffff16815260200191505060405180910390f35b6060600180546001816001' +
            '16156101000203166002900480601f0160208091040260200160405190810160' +
            '4052809291908181526020018280546001816001161561010002031660029004' +
            '801561020c5780601f106101e15761010080835404028352916020019161020c' +
            '565b820191906000526020600020905b81548152906001019060200180831161' +
            '01ef57829003601f168201915b5050505050905090565b600080905490610100' +
            '0a900473ffffffffffffffffffffffffffffffffffffffff1673ffffffffffff' +
            'ffffffffffffffffffffffffffff163373ffffffffffffffffffffffffffffff' +
            'ffffffffff161461026f57600080fd5b806000806101000a81548173ffffffff' +
            'ffffffffffffffffffffffffffffffff021916908373ffffffffffffffffffff' +
            'ffffffffffffffffffff16021790555050565b60008060009054906101000a90' +
            '0473ffffffffffffffffffffffffffffffffffffffff1690509056fea165627a' +
            '7a723058203d58598dee9113d96d02a8821205cf2af6e04d52d4d3b52f4563f7' +
            '7936acb5490029')
    storage = {hex(0): str(Web3.toChecksumAddress(owner_address))}
    storage = {**storage, **_string_to_storage(1, schain_name)}

    return code, storage


def _string_to_storage(slot: int, string: str) -> dict:
    # https://solidity.readthedocs.io/en/develop/miscellaneous.html#bytes-and-string
    storage = dict()
    binary = string.encode()
    length = len(binary)
    if length < 32:
        binary += (2 * length).to_bytes(32 - length, 'big')
        storage[hex(slot)] = '0x' + binary.hex()
    else:
        storage[hex(slot)] = hex(2 * length + 1)

        keccak_hash = keccak.new(digest_bits=256)
        keccak_hash.update(slot.to_bytes(32, 'big'))
        offset = int.from_bytes(keccak_hash.digest(), 'big')

        def chunks(size, source):
            for i in range(0, len(source), size):
                yield source[i:i + size]

        for index, data in enumerate(chunks(32, binary)):
            if len(data) < 32:
                data += int(0).to_bytes(32 - len(data), 'big')
            storage[hex(offset + index)] = '0x' + data.hex()
    return storage


def add_chain_id(base_config, schain_name):
    keccak_hash = keccak.new(digest_bits=256)
    keccak_hash.update(schain_name.encode("utf-8"))
    hash_ = keccak_hash.hexdigest()
    hash_ = hash_[:13]			# use 52 bits
    base_config['params']['chainID'] = "0x" + hash_



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


class TransactionFailedError(Exception):
    pass


class TxRes():
    def __init__(self, tx_hash=None, data=None, receipt=None):
        self._hash = tx_hash
        self._data = data
        self._receipt = receipt

    @property
    def hash(self):
        return self._hash

    @property
    def data(self):
        return self._data

    @property
    def receipt(self):
        return self._receipt

    @receipt.setter
    def receipt(self, receipt: dict) -> None:
        self._receipt = receipt

    def raise_for_status(self) -> None:
        if self._receipt and self._receipt['status'] != 1:
            raise TransactionFailedError(f'Transaction failed with receipt '
                                         f'{self._receipt}')



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import skale.contracts as contracts
from skale.utils.contract_info import ContractInfo
from skale.utils.contract_types import ContractTypes

CONTRACTS_INFO = [
    ContractInfo('contract_manager', 'ContractManager',
                 contracts.ContractManager, ContractTypes.API, False),
    ContractInfo('token', 'SkaleToken', contracts.Token, ContractTypes.API,
                 False),
    ContractInfo('manager', 'SkaleManager', contracts.Manager,
                 ContractTypes.API, True),
    ContractInfo('constants_holder', 'ConstantsHolder', contracts.ConstantsHolder,
                 ContractTypes.INTERNAL, True),
    ContractInfo('nodes', 'NodesFunctionality', contracts.Nodes,
                 ContractTypes.API, True),
    ContractInfo('schains', 'SchainsFunctionality', contracts.SChains,
                 ContractTypes.API, True),
    ContractInfo('monitors', 'MonitorsFunctionality', contracts.MonitorsFunctionality,
                 ContractTypes.API, True),
    ContractInfo('nodes_data', 'NodesData', contracts.NodesData,
                 ContractTypes.DATA, True),
    ContractInfo('schains_data', 'SchainsData', contracts.SChainsData,
                 ContractTypes.DATA, True),
    ContractInfo('monitors_data', 'MonitorsData', contracts.MonitorsData,
                 ContractTypes.DATA, True),
    ContractInfo('dkg', 'SkaleDKG', contracts.DKG, ContractTypes.API, True),
    ContractInfo('delegation_controller', 'DelegationController', contracts.DelegationController,
                 ContractTypes.API, False),
    ContractInfo('validator_service', 'ValidatorService', contracts.ValidatorService,
                 ContractTypes.API, False),
    ContractInfo('token_state', 'TokenState', contracts.TokenState,
                 ContractTypes.API, False),
    ContractInfo('distributor', 'Distributor', contracts.Distributor,
                 ContractTypes.API, False)
]


def get_contracts_info():
    contracts_info = {}
    for contract_info in CONTRACTS_INFO:
        contracts_info[contract_info.name] = contract_info
    return contracts_info
